@@ -28,12 +28,8 @@
 #include <string.h>
 #include <assert.h>
 #include <ctype.h>
-#include <signal.h>
 #include <limits.h>
 #include <sys/types.h>
-#ifdef HAVE_SYS_SYSCTL_H
-# include <sys/sysctl.h>
-#endif
 
 #include "widl.h"
 #include "utils.h"
@@ -46,6 +42,7 @@ static const char usage[] =
 "Usage: widl [options...] infile.idl\n"
 "   or: widl [options...] --dlldata-only name1 [name2...]\n"
 "   --acf=file         Use ACF file\n"
+"   --align=n          Set structure packing to 'n'\n"
 "   -app_config        Ignored, present for midl compatibility\n"
 "   -b arch            Set the target architecture\n"
 "   -c                 Generate client stub\n"
@@ -63,9 +60,11 @@ static const char usage[] =
 "   --nostdinc         Do not search the standard include path\n"
 "   --ns_prefix        Prefix namespaces with ABI namespace\n"
 "   --oldnames         Use old naming conventions\n"
+"   --oldtlb           Generate typelib in the old format (SLTG)\n"
 "   -o, --output=NAME  Set the output file name\n"
 "   -Otype             Type of stubs to generate (-Os, -Oi, -Oif)\n"
 "   -p                 Generate proxy\n"
+"   --packing=n        Set structure packing to 'n'\n"
 "   --prefix-all=p     Prefix names of client stubs / server functions with 'p'\n"
 "   --prefix-client=p  Prefix names of client stubs with 'p'\n"
 "   --prefix-server=p  Prefix names of server functions with 'p'\n"
@@ -78,8 +77,6 @@ static const char usage[] =
 "   -V                 Print version and exit\n"
 "   -W                 Enable pedantic warnings\n"
 "   --win32, --win64   Set the target architecture (Win32 or Win64)\n"
-"   --win32-align n    Set win32 structure alignment to 'n'\n"
-"   --win64-align n    Set win64 structure alignment to 'n'\n"
 "   --winrt            Enable Windows Runtime mode\n"
 "Debug level 'n' is a bitmask with following meaning:\n"
 "    * 0x01 Tell which resource is parsed (verbose mode)\n"
@@ -93,7 +90,7 @@ static const char usage[] =
 static const char version_string[] = "Wine IDL Compiler version " PACKAGE_VERSION "\n"
 			"Copyright 2002 Ove Kaaven\n";
 
-static struct target target;
+struct target target = { 0 };
 
 int debuglevel = DEBUGLEVEL_NONE;
 int parser_debug, yy_flex_debug;
@@ -111,15 +108,14 @@ int do_idfile = 0;
 int do_dlldata = 0;
 static int no_preprocess = 0;
 int old_names = 0;
-int win32_packing = 8;
-int win64_packing = 8;
+int old_typelib = 0;
 int winrt_mode = 0;
+int interpreted_mode = 1;
 int use_abi_namespace = 0;
 static int stdinc = 1;
-static enum stub_mode stub_mode = MODE_Os;
 
 char *input_name;
-char *input_idl_name;
+char *idl_name;
 char *acf_name;
 char *header_name;
 char *local_stubs_name;
@@ -135,7 +131,8 @@ char *server_token;
 char *regscript_name;
 char *regscript_token;
 static char *idfile_name;
-char *temp_name;
+struct strarray temp_files = { 0 };
+const char *temp_dir = NULL;
 const char *prefix_client = "";
 const char *prefix_server = "";
 static const char *includedir;
@@ -144,10 +141,9 @@ static struct strarray dlldirs;
 static char *output_name;
 static const char *sysroot = "";
 
-int line_number = 1;
-
 static FILE *idfile;
 
+unsigned int packing = 8;
 unsigned int pointer_size = 0;
 
 time_t now;
@@ -160,6 +156,8 @@ enum {
     DLLDATA_ONLY_OPTION,
     LOCAL_STUBS_OPTION,
     NOSTDINC_OPTION,
+    OLD_TYPELIB_OPTION,
+    PACKING_OPTION,
     PREFIX_ALL_OPTION,
     PREFIX_CLIENT_OPTION,
     PREFIX_SERVER_OPTION,
@@ -170,14 +168,13 @@ enum {
     SYSROOT_OPTION,
     WIN32_OPTION,
     WIN64_OPTION,
-    WIN32_ALIGN_OPTION,
-    WIN64_ALIGN_OPTION
 };
 
 static const char short_options[] =
     "b:cC:d:D:EhH:I:L:m:No:O:pP:rsS:tT:uU:VW";
 static const struct long_option long_options[] = {
     { "acf", 1, ACF_OPTION },
+    { "align", 1, PACKING_OPTION },
     { "app_config", 0, APP_CONFIG_OPTION },
     { "dlldata", 1, DLLDATA_OPTION },
     { "dlldata-only", 0, DLLDATA_ONLY_OPTION },
@@ -186,7 +183,9 @@ static const struct long_option long_options[] = {
     { "nostdinc", 0, NOSTDINC_OPTION },
     { "ns_prefix", 0, RT_NS_PREFIX },
     { "oldnames", 0, OLDNAMES_OPTION },
+    { "oldtlb", 0, OLD_TYPELIB_OPTION },
     { "output", 0, 'o' },
+    { "packing", 1, PACKING_OPTION },
     { "prefix-all", 1, PREFIX_ALL_OPTION },
     { "prefix-client", 1, PREFIX_CLIENT_OPTION },
     { "prefix-server", 1, PREFIX_SERVER_OPTION },
@@ -196,20 +195,11 @@ static const struct long_option long_options[] = {
     { "winrt", 0, RT_OPTION },
     { "win32", 0, WIN32_OPTION },
     { "win64", 0, WIN64_OPTION },
-    { "win32-align", 1, WIN32_ALIGN_OPTION },
-    { "win64-align", 1, WIN64_ALIGN_OPTION },
     { NULL }
 };
 
 static void rm_tempfile(void);
 
-enum stub_mode get_stub_mode(void)
-{
-    /* old-style interpreted stubs are not supported on 64-bit */
-    if (stub_mode == MODE_Oi && pointer_size == 8) return MODE_Oif;
-    return stub_mode;
-}
-
 static char *make_token(const char *name)
 {
   char *token;
@@ -441,7 +431,7 @@ void write_id_data(const statement_list_t *stmts)
   }
 
   fprintf(idfile, "/*** Autogenerated by WIDL %s ", PACKAGE_VERSION);
-  fprintf(idfile, "from %s - Do not edit ***/\n\n", input_idl_name);
+  fprintf(idfile, "from %s - Do not edit ***/\n\n", idl_name);
   fprintf(idfile, "#include <rpc.h>\n");
   fprintf(idfile, "#include <rpcndr.h>\n\n");
 
@@ -482,25 +472,11 @@ void write_id_data(const statement_list_t *stmts)
 
 static void init_argv0_dir( const char *argv0 )
 {
-#ifndef _WIN32
-    char *dir;
+    char *dir = get_argv0_dir( argv0 );
 
-#if defined(__linux__) || defined(__FreeBSD_kernel__) || defined(__NetBSD__)
-    dir = realpath( "/proc/self/exe", NULL );
-#elif defined (__FreeBSD__) || defined(__DragonFly__)
-    static int pathname[] = { CTL_KERN, KERN_PROC, KERN_PROC_PATHNAME, -1 };
-    size_t path_size = PATH_MAX;
-    char *path = malloc( path_size );
-    if (path && !sysctl( pathname, sizeof(pathname)/sizeof(pathname[0]), path, &path_size, NULL, 0 ))
-        dir = realpath( path, NULL );
-    free( path );
-#else
-    dir = realpath( argv0, NULL );
-#endif
     if (!dir) return;
-    includedir = strmake( "%s/%s", get_dirname( dir ), BIN_TO_INCLUDEDIR );
-    dlldir = strmake( "%s/%s", get_dirname( dir ), BIN_TO_DLLDIR );
-#endif
+    includedir = strmake( "%s/%s", dir, BIN_TO_INCLUDEDIR );
+    dlldir = strmake( "%s/%s", dir, BIN_TO_DLLDIR );
 }
 
 static void option_callback( int optc, char *optarg )
@@ -552,15 +528,10 @@ static void option_callback( int optc, char *optarg )
     case WIN64_OPTION:
       pointer_size = 8;
       break;
-    case WIN32_ALIGN_OPTION:
-      win32_packing = strtol(optarg, NULL, 0);
-      if(win32_packing != 2 && win32_packing != 4 && win32_packing != 8)
-          error("Packing must be one of 2, 4 or 8\n");
-      break;
-    case WIN64_ALIGN_OPTION:
-      win64_packing = strtol(optarg, NULL, 0);
-      if(win64_packing != 2 && win64_packing != 4 && win64_packing != 8)
-          error("Packing must be one of 2, 4 or 8\n");
+    case PACKING_OPTION:
+      packing = strtol(optarg, NULL, 0);
+      if(packing != 2 && packing != 4 && packing != 8)
+          error("Structure packing must be one of 2, 4 or 8\n");
       break;
     case ACF_OPTION:
       acf_name = xstrdup(optarg);
@@ -617,11 +588,11 @@ static void option_callback( int optc, char *optarg )
       output_name = xstrdup(optarg);
       break;
     case 'O':
-      if (!strcmp( optarg, "s" )) stub_mode = MODE_Os;
-      else if (!strcmp( optarg, "i" )) stub_mode = MODE_Oi;
-      else if (!strcmp( optarg, "ic" )) stub_mode = MODE_Oif;
-      else if (!strcmp( optarg, "if" )) stub_mode = MODE_Oif;
-      else if (!strcmp( optarg, "icf" )) stub_mode = MODE_Oif;
+      if (!strcmp( optarg, "s" )) interpreted_mode = 0;
+      else if (!strcmp( optarg, "i" )) interpreted_mode = 1;
+      else if (!strcmp( optarg, "ic" )) interpreted_mode = 1;
+      else if (!strcmp( optarg, "if" )) interpreted_mode = 1;
+      else if (!strcmp( optarg, "icf" )) interpreted_mode = 1;
       else error( "Invalid argument '-O%s'\n", optarg );
       break;
     case 'p':
@@ -646,6 +617,9 @@ static void option_callback( int optc, char *optarg )
       do_everything = 0;
       do_typelib = 1;
       break;
+    case OLD_TYPELIB_OPTION:
+      old_typelib = 1;
+      break;
     case 'T':
       typelib_name = xstrdup(optarg);
       break;
@@ -687,8 +661,8 @@ int open_typelib( const char *name )
         {
             int namelen = strlen( name );
             if (strendswith( name, ".dll" )) namelen -= 4;
-            TRYOPEN( strmake( "%.*s/%.*s/%s", (int)strlen(dlldirs.str[i]) - 2, dlldirs.str[i],
-                              namelen, name, name ));
+            TRYOPEN( strmake( "%.*s/%.*s%s/%s", (int)strlen(dlldirs.str[i]) - 2, dlldirs.str[i],
+                              namelen, name, pe_dir, name ));
         }
         else
         {
@@ -710,7 +684,7 @@ int open_typelib( const char *name )
             TRYOPEN( strmake( "%s%s/%s", default_dirs[i], pe_dir, name ));
         }
     }
-    return -1;
+    error( "cannot find %s\n", name );
 #undef TRYOPEN
 }
 
@@ -720,11 +694,7 @@ int main(int argc,char *argv[])
   int ret = 0;
   struct strarray files;
 
-  signal( SIGTERM, exit_on_signal );
-  signal( SIGINT, exit_on_signal );
-#ifdef SIGHUP
-  signal( SIGHUP, exit_on_signal );
-#endif
+  init_signals( exit_on_signal );
   init_argv0_dir( argv[0] );
   target = init_argv0_target( argv[0] );
 
@@ -799,7 +769,10 @@ int main(int argc,char *argv[])
       return 1;
     }
     else
-      input_idl_name = input_name = xstrdup(files.str[0]);
+    {
+      input_name = xstrdup( files.str[0] );
+      idl_name = get_basename( input_name );
+    }
   }
   else {
     fprintf(stderr, "%s", usage);
@@ -849,53 +822,15 @@ int main(int argc,char *argv[])
   wpp_add_cmdline_define("_WIN32=1");
 
   atexit(rm_tempfile);
-  if (!no_preprocess)
-  {
-    chat("Starting preprocess\n");
-
-    if (!preprocess_only)
-    {
-        FILE *output;
-        int fd;
-        char *name;
-
-        fd = make_temp_file( header_name, NULL, &name );
-        temp_name = name;
-        if (!(output = fdopen(fd, "wt")))
-            error("Could not open fd %s for writing\n", name);
-
-        ret = wpp_parse( input_name, output );
-        fclose( output );
-    }
-    else
-    {
-        ret = wpp_parse( input_name, stdout );
-    }
-
-    if(ret) exit(1);
-    if(preprocess_only) exit(0);
-    if(!(parser_in = fopen(temp_name, "r"))) {
-      fprintf(stderr, "Could not open %s for input\n", temp_name);
-      return 1;
-    }
-  }
-  else {
-    if(!(parser_in = fopen(input_name, "r"))) {
-      fprintf(stderr, "Could not open %s for input\n", input_name);
-      return 1;
-    }
-  }
+  if (preprocess_only) exit( wpp_parse( input_name, stdout ) );
+  parser_in = open_input_file( input_name );
 
   header_token = make_token(header_name);
 
   init_types();
   ret = parser_parse();
-
-  fclose(parser_in);
-
-  if(ret) {
-    exit(1);
-  }
+  close_all_inputs();
+  if (ret) exit(1);
 
   /* Everything has been done successfully, don't delete any files.  */
   set_everything(FALSE);
@@ -906,9 +841,6 @@ int main(int argc,char *argv[])
 
 static void rm_tempfile(void)
 {
-  abort_import();
-  if(temp_name)
-    unlink(temp_name);
   if (do_header)
     unlink(header_name);
   if (local_stubs_name)
@@ -925,4 +857,38 @@ static void rm_tempfile(void)
     unlink(proxy_name);
   if (do_typelib)
     unlink(typelib_name);
+  remove_temp_files();
+}
+
+char *find_input_file( const char *name, const char *parent )
+{
+    char *path;
+
+    /* don't search for a file name with a path in the include directories, for compatibility with MIDL */
+    if (strchr( name, '/' ) || strchr( name, '\\' )) path = xstrdup( name );
+    else if (!(path = wpp_find_include( name, parent ))) error_loc( "Unable to open include file %s\n", name );
+
+    return path;
+}
+
+FILE *open_input_file( const char *path )
+{
+    FILE *file;
+    char *name;
+    int ret;
+
+    if (no_preprocess)
+    {
+        if (!(file = fopen( path, "r" ))) error_loc( "Unable to open %s\n", path );
+        return file;
+    }
+
+    name = make_temp_file( "widl", NULL );
+    if (!(file = fopen( name, "wt" ))) error_loc( "Could not open %s for writing\n", name );
+    ret = wpp_parse( path, file );
+    fclose( file );
+    if (ret) exit( 1 );
+
+    if (!(file = fopen( name, "r" ))) error_loc( "Unable to open %s\n", name );
+    return file;
 }
